/* A bit o' ballistics,
   by Chusslove Illich <caslav.ilic@gmx.net>, 2007-02-03 */

#include <stdio.h>
#include <math.h>

/* x squared. */
#define P2(x) ((x)*(x))

/* x to radians. */
#define RAD(x) ((x)*M_PI/180)

/* x to degrees. */
#define DEG(x) ((x)*180/M_PI)

int main ()
{
    /* Shot start is at (x, y) coordinate center origin (0, 0),
       x being the horizontal axis in direction of the target,
       y vertical upwards,
       range computation to target at level altitude. */

    /* Deceleration due to air resistance (drag) is modeled as
       proportional to square of velocity.
       Don't forget gravity acceleration (g). */

    /* Downward acceleration by gravity [m s^-2]. */
    double g = 9.81;

    /* Proportionality factor in drag model [m^-1]. */
    double k = 2.20e-5;
    /* Muzzle velocity [m s^-1]. */
    double vel0 = 732.0;
    /* Muzzle elevations [deg]. */
    double a0[] = {2.6, 5.6, 9.3, 13.8, 19.2, 26.1, 30.5};

    /* Allowed integration error: max range delta at target [m]. */
    double xdelt = 10.0;

    /* On trajectory quantities: */
    double x, y; /* position [m] */
    double u, v; /* velocity [m s^-1] */
    double a; /* trajectory angle [rad] */
    double t; /* flight time [s] */

    int na0 = sizeof(a0) / sizeof(double); /* number of elevations */
    double dt; /* flight time step size */
    int i; /* elevation loop counter */

    /* Output fixed data. */
    printf("Drag constant [m^-1]: %.2e\n", k);
    printf("Muzzle velocity [m s^-1]: %.1f\n", vel0);
    printf("Max range delta at target [m]: %.1f\n", xdelt);
    printf("Solutions per muzzle elevation:\n");
    printf("%10s, %10s, %10s, %10s, %10s, %10s\n",
           "elevation", "range", "ftime", "impactvel", "impactang", "tstep");
    printf("%10s, %10s, %10s, %10s, %10s, %10s\n",
           "[deg]", "[m]", "[s]", "[m s^-1]", "[deg]", "[s]");

    /* Compute for all elevations. */
    for (i = 0; i < na0; ++i)
    {
        dt = 0.1; /* initial time step size */

        /* Adapt step size for precise enough integration */
        while (1)
        {
            /* Initial trajectory values. */
            x = 0.0;
            y = 0.0;
            u = vel0 * cos(RAD(a0[i]));
            v = vel0 * sin(RAD(a0[i]));
            a = a0[i];
            t = 0.0;

            /* Time integration, up to when the shot dips below level. */
            while (y >= 0.0)
            {
                double dacc; /* deceleration due to drag. */
                double pn, qn; /* new acceleration components */
                double xn, yn; /* new position components */
                double un, vn; /* new velocity components */
                double an; /* new trajectory angle. */

                /* New trajectory angle. */
                an = DEG(atan2(v, u));

                /* New drag deceleration. */
                dacc = k * (P2(u) + P2(v));

                /* New acceleration components. */
                pn = -dacc * cos(RAD(an));
                qn = -dacc * sin(RAD(an)) - g;

                /* New velocity components. */
                un = u + pn * dt;
                vn = v + qn * dt;

                /* New position. */
                xn = x + un * dt;
                yn = y + vn * dt;

                /* Make sure integration is still stable,
                otherwise abort and force step size reduction. */
                if (   (un > u) /* cannot speed up horizontaly */
                    || (vn > v) /* cannot speed up verticaly */
                    || (xn < x) /* cannot go backwards */
                )
                {
                    /* Force step size reduction in the post loop check. */
                    v = 1.1 * xdelt / dt;
                    break;
                }

                /* Copy over values for next step. */
                x = xn; y = yn;
                u = un; v = vn;
                a = an;

                /* Update flight time. */
                t += dt;
            }

            /* Check if result is precise enough. */
            if (dt * u > xdelt)
                /* Reduce step size if shot too much below target. */
                dt *= 0.5;
            else
                /* We're done: x contains the range. */
                break;
        }

        /* Output solution for this elevation. */
        printf("%10.1f, %10.0f, %10.1f, %10.0f, %10.1f, %10.5f\n",
            a0[i], x, t, sqrt(P2(u) + P2(v)), a, dt);
    }

    return 0;
}
